/******************************************************************************/
/*                                                                            */
/*                        iMX1/L IDE DEVICE MODULE V 2.0                      */
/*                        2005 FREESCALE SEMICONDUCTOR                        */             
/*                                                                            */
/* This module allows to plug an ATA hard drive device						  */
/* into iMX1/L ATA daugterboard.                                                                     */
/*                                                                            */
/* Authors: Erwan Bracq, Florent Auger                                        */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*                                VERSIONS                                    */
/*                                                                            */
/* 0.1 ALPHA initial release                                                  */
/* 0.2 ALPHA write / read sector attempt [ok, write read possible]            */
/* 0.3 ALPHA interrupt attempt [ok, interrupt avalaible]                      */
/* 0.4 ALPHA block device structure                                           */
/* 0.5 ALPHA interrupt structure                                              */
/* 0.6 ALPHA debug                                                            */
/* 0.7 ALPHA software clustered requests 512/1024                             */
/*     (with hardware specific command)                                       */
/* 0.8 ALPHA PCB adpated driver                                               */
/* 0.9 ALPHA modify masks applied for GPIO Port A IRQ                         */
/*     (conflict with ethernet)                                               */
/* 1.0 ALPHA remove CSCR programming and change WSC for 96MHz access          */
/* 1.1 ALPHA debug mx1_IDE_wait_signal function                               */
/* 2.0 ALPHA implement DMA for Read and Write transfers
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*                                COMMENTS                                    */
/*                                                                            */
/* DMA is configured out of spec to avoid back-to-back read issue, cause the  */
/* HDD does not seem to support it. It works if ONLY ONE DMA channel is used  */
/* at a time. Conflict with SSI driver for instance if not.					  */
/* It is not recommended to use it.											  */
/*																			  */
/******************************************************************************/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

/* kernel and module includes */
#include <linux/config.h>

/* checks kernel version */
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#include <linux/module.h>
#include <linux/init.h>

#define MAJOR_NR major                                       /* force definitions on in blk.h */
int major;                                                   /* must be declared before including blk.h */

#define MX1_IDE_SHIFT 6                                      /* 64 partitions per device */
#define MX1_IDE_NPART (1 << MX1_IDE_SHIFT)                   /* number of partitions */

#define DEVICE_NR(device) (MINOR(device) >> MX1_IDE_SHIFT)   /* 64 partitions avalaible */
#define DEVICE_NAME "hd"                                     /* device name */
#define DEVICE_NO_RANDOM                                     /* no entropy to contribute */
#define DEVICE_REQUEST mx1_IDE_request                       /* request routine */  
#define REALLY_SLOW_IO

/* block device functions */
#include <linux/blk.h>
#include <linux/hdreg.h>
#include <linux/blkpg.h>

/* memory allocations */
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include <linux/interrupt.h>
#include <linux/delay.h>

/* direct memory access */
#ifdef DMA_MODE
#include <asm/dma.h>
#include <asm/arch/dma.h>
#include <asm/mach/dma.h>
#endif

/* specific includes */
#include "mx1_IDE.h"

/* IDE drive internal representation */
static ide_dev      *mx1_ide_dev = NULL; 
static ide_dev_id    mx1_ide_dev_id; 

/* block device global parameters */ 
static int    ide_major = MX1_IDE_MAJOR;                   /* major number */
static int    ide_readahead = MX1_IDE_READAHEAD;           /* read ahead, 8 sectors */
static int    ide_hardsect_size[MX1_IDE_NPART];            /* used for hardsect_size - should be 512 bytes */
static int    ide_maxs[MX1_IDE_NPART];                     /* used for max_sectors[] - limit size of individual request */
static int    ide_blk_size[MX1_IDE_NPART];                 /* used in gendisk - gives whole size of partition */
static int    ide_blksize_size[MX1_IDE_NPART];             /* logical sector size */
static struct hd_struct ide_partitions[MX1_IDE_NPART];     /* used in gendisk - gives particular partition information */

/* forward declaration */
static struct block_device_operations mx1_IDE_bdops;       /* block device operations */
static void mx1_IDE_reset();                               /* reset device function */
static void mx1_IDE_wait_signal(int, int);                 /* active wait on status register */
static void mx1_IDE_transtask();                           /* request function */
static void mx1_IDE_rtask();                               /* read on interrupt tasklet */
static void mx1_IDE_wtask();                               /* write tasklet */
static void mx1_IDE_errtask();                             /* error tasklet */

/* tasklet declaration */
DECLARE_TASKLET (mx1_IDE_read_tasklet, mx1_IDE_rtask, 0); 
DECLARE_TASKLET (mx1_IDE_write_tasklet, mx1_IDE_wtask, 0);
DECLARE_TASKLET (mx1_IDE_error_tasklet, mx1_IDE_errtask, 0);
DECLARE_TASKLET (mx1_IDE_transfer_tasklet, mx1_IDE_transtask, 0);

/* generic hard drive structure, partition support */
static struct gendisk mx1_ide_gendisk = {
  major:              0,                                   /* major number assigned later */
  major_name:         DEVICE_NAME,                         /* name of the major device */
  minor_shift:        MX1_IDE_SHIFT,                       /* shift to get device number */
  max_p:              MX1_IDE_NPART,                       /* number of partitions */      
  part:               ide_partitions,                      /* partitions infos */
  sizes:              ide_blk_size,                        /* partitions sizes */
  real_devices:       (void *)&mx1_ide_dev_id,             /* device extra infos */
  fops:               &mx1_IDE_bdops,                      /* block dev operations */          
};
static char   mx1_ide_gendisk_flags;

/* DMA specific */
#ifdef DMA_MODE
static dmach_t mx1ide_dma_chan_rd;
static void mx1ide_dma_intr_handler_rd(void);
static void mx1ide_dma_err_handler_rd(int error_type);
static dmach_t mx1ide_dma_chan_wr;
static void mx1ide_dma_intr_handler_wr(void);
static void mx1ide_dma_err_handler_wr(int error_type);

#define mx1_dma_chan_reg_out(r,c,x) outl(x,IO_ADDRESS((DMA_CH0_BASE + r + c * DMA_REG_SET_OFS * sizeof(VU32))))

extern void mx1ads_request_dma_intr(dmach_t channel,
				    callback_t dma_callback,
				    err_callback_t dma_err_callback);
extern void mx1ads_free_dma_intr(dmach_t channel);
#endif

/******************************************************************************/
/*                                                                            */
/*                      MALLOC & FREE                                         */
/*                                                                            */
/******************************************************************************/
static int mx1_IDE_malloc(void)
     /* allocate memory for IDE device driver, GPIO and EIM */
{
  printk(KERN_INFO "Requesting iMX1/L E/S regions... ");

  /* probe I/O regions */
  /* commons GPIO pins 2, 17, 18 */
  /* pin 2 drive LED2 */
  /* pin 17 drive reset signal */
  /* pin 18 drive interrupt signal */
  if(check_region(PTA_GIUS_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTA_OCR1_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTA_ICONFA2_, sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTA_DDIR_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTA_PUEN_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTA_DR_     , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTA_ICR2_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTA_ISR_    , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTA_IMR_    , sizeof(U32)) < 0) return -ENOMEM;
  
  if(check_region(PTB_GIUS_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTB_OCR2_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTB_DDIR_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTB_PUEN_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(PTB_DR_     , sizeof(U32)) < 0) return -ENOMEM;
  
  /* EIM */
  /* enables CS5 */
  if(check_region(EIM_CS5H_   , sizeof(U32)) < 0) return -ENOMEM;
  if(check_region(EIM_CS5L_   , sizeof(U32)) < 0) return -ENOMEM;

  if(check_region(CRM_CSCR_   , sizeof(U32)) < 0) return -ENOMEM;

  /* IDE */
  if(check_region(IDE_MEM_BASE, IDE_MEM_DIM*sizeof(U8)) < 0) return -ENOMEM;

  /* requests I/O regions */
  /* commons GPIO pins 2, 17, 18 */
  request_mem_region(PTA_GIUS_   , sizeof(U32), "reserved");
  request_mem_region(PTA_OCR1_   , sizeof(U32), "reserved");
  request_mem_region(PTA_ICONFA2_, sizeof(U32), "reserved");
  request_mem_region(PTA_DDIR_   , sizeof(U32), "reserved");
  request_mem_region(PTA_PUEN_   , sizeof(U32), "reserved");
  request_mem_region(PTA_DR_     , sizeof(U32), "reserved");
  request_mem_region(PTA_ICR2_   , sizeof(U32), "reserved");
  request_mem_region(PTA_ISR_    , sizeof(U32), "reserved");
  request_mem_region(PTA_IMR_    , sizeof(U32), "reserved");

  request_mem_region(PTB_GIUS_   , sizeof(U32), "reserved");
  request_mem_region(PTB_OCR2_   , sizeof(U32), "reserved");
  request_mem_region(PTB_DDIR_   , sizeof(U32), "reserved");
  request_mem_region(PTB_PUEN_   , sizeof(U32), "reserved");
  request_mem_region(PTB_DR_     , sizeof(U32), "reserved");

  /* EIM */
  request_mem_region(EIM_CS5H_   , sizeof(U32), "reserved");
  request_mem_region(EIM_CS5L_   , sizeof(U32), "reserved");

  /* CLOCK */
  request_mem_region(CRM_CSCR_   , sizeof(U32), "reserved");

  /* IDE */
  request_mem_region(IDE_MEM_BASE, IDE_MEM_DIM, "ide_device");
 
  return 0;
}

static void mx1_IDE_free(void)
     /* release all memory allocations */
{ 
  printk(KERN_INFO "Releasing iMX1/L E/S regions... ");
 
  /* releases I/O regions */
  /* commons GPIO pins 2, 17, 18 */
  release_mem_region(PTA_GIUS_   , sizeof(U32));
  release_mem_region(PTA_OCR1_   , sizeof(U32));
  release_mem_region(PTA_ICONFA2_, sizeof(U32));
  release_mem_region(PTA_DDIR_   , sizeof(U32));
  release_mem_region(PTA_PUEN_   , sizeof(U32));
  release_mem_region(PTA_DR_     , sizeof(U32));
  release_mem_region(PTA_ICR2_   , sizeof(U32));
  release_mem_region(PTA_ISR_    , sizeof(U32));
  release_mem_region(PTA_IMR_    , sizeof(U32));
  
  release_mem_region(PTB_GIUS_   , sizeof(U32));
  release_mem_region(PTB_OCR2_   , sizeof(U32));
  release_mem_region(PTB_DDIR_   , sizeof(U32));
  release_mem_region(PTB_PUEN_   , sizeof(U32));
  release_mem_region(PTB_DR_     , sizeof(U32));
   
  /* EIM */
  release_mem_region(EIM_CS5H_   , sizeof(U32));
  release_mem_region(EIM_CS5L_   , sizeof(U32));
  
  release_mem_region(CRM_CSCR_   , sizeof(U32));

  /* IDE */
  release_mem_region(IDE_MEM_BASE, IDE_MEM_DIM*sizeof(U8));
  
  printk(KERN_INFO "completed\n");
}

static void mx1_IDE_make_clean(void)
     /* free all allocated memories */
{
  /* stop internal ide structure */
  spin_lock(&mx1_ide_dev->lock_dev);
  mx1_ide_dev->usage++;
  spin_unlock(&mx1_ide_dev->lock_dev);  
  /* clean internal ide structure */
  kfree(mx1_ide_dev);
  mx1_IDE_free();

#ifdef DMA_MODE
  /* free DMA channels */
	if (mx1ide_dma_chan_rd != -1) {
		mx1ads_free_dma_intr(mx1ide_dma_chan_rd);
		free_dma(mx1ide_dma_chan_rd);
	}
	if (mx1ide_dma_chan_wr != -1) {
		mx1ads_free_dma_intr(mx1ide_dma_chan_wr);
		free_dma(mx1ide_dma_chan_wr);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/*                      FUNCTION TO CONFIGURE EIM AND CLOCK                   */
/*                                                                            */
/******************************************************************************/
static void mx1_EIM_init(void)
     /* configure EIM */
{
//#define AGGRESSIVE
#ifdef AGGRESSIVE
  /* enables CS5 */
  outl(0xa00, (volatile P_U32) EIM_CS5H_); /* wait states */
  /* EIM time cycles */
  outl(0x52520d01, (volatile P_U32) EIM_CS5L_); /* OE and EB3 timings configuration */
#else
  /* enables CS5 */
  outl(0xb00, (volatile P_U32) EIM_CS5H_); /* wait states */
  /* EIM time cycles */
  outl(0x63630d01, (volatile P_U32) EIM_CS5L_); /* OE and EB3 timings configuration */
#endif
}


/******************************************************************************/
/*                                                                            */
/*                      FUNCTION TO CONFIGURE GPIO                            */
/*                                                                            */
/******************************************************************************/
static void mx1_GPIO_init(void)
     /* configure GPIO pins */
{
  /* GPIO common operations */
  *(volatile P_U32) (PTA_GIUS_) |=  0x40004;  /* pins 2, 18 used for GPIO function */

  /* GPIO pin 2, LED2 driver */
  *(volatile P_U32) (PTA_OCR1_) |=  0x30;     /* pin 2 is a data register output signal */
  *(volatile P_U32) (PTA_DDIR_) |=  0x4;      /* pin 2 is an output signal */
 
  /* GPIO pin 17, reset signal */
  *(volatile P_U32) (PTB_GIUS_) |=  0x20000;  /* pin 17 used for GPIO function */
  *(volatile P_U32) (PTB_OCR2_) |=  0xc;      /* pin 17 is a data register output signal */
  *(volatile P_U32) (PTB_DDIR_) |=  0x20000;  /* pin 17 is an output signal */
  *(volatile P_U32) (PTB_PUEN_) &= ~0x20000;  /* pin 17 is tri-stated when not driven */
  *(volatile P_U32) (PTB_DR_)   &= ~0x20000;  /* drives the pin 17 output signal low */

  /* GPIO pin 18, interrupt signal IRQ 11 */
  *(volatile P_U32) (PTA_DDIR_)    &= ~0x40000;  /* pin 18 is an input signal */ 
  *(volatile P_U32) (PTA_ICONFA2_) &= ~0x30;     /* pin 18 is an GPIO in */
  *(volatile P_U32) (PTA_PUEN_)    &= ~0x40000;  /* pin 18 is tri-stated when not driven */
  *(volatile P_U32) (PTA_ICR2_)    &= ~0x30;     /* configure GPIO pin 18 for negative edge sensitive */
  *(volatile P_U32) (PTA_ICR2_)    |=  0x10;     /* configure GPIO pin 18 for negative edge sensitive */
}

/******************************************************************************/
/*                                                                            */
/*                      HARDWARE INITIALISATION                               */
/*                                                                            */
/******************************************************************************/
static int mx1_IDE_hard_init(void)
     /* hardware initialization */
{
  int result = 0;
  
  /* step 1 request memory E/S */
  result = mx1_IDE_malloc();
  if(result == 0) printk(KERN_INFO "success\n");else {
    printk(KERN_INFO "fail\n"); 
    mx1_IDE_make_clean();
    return -ENOMEM; /* out of memory */
  }
  
  /* step 2 configure MX1 GPIO */
  printk(KERN_INFO "Configure iMX1/L GPIO.\n");
  mx1_GPIO_init();
 
  /* step 3 configure MX1 EIM */
  printk(KERN_INFO "Configure iMX1/L EIM.\n");
  mx1_EIM_init();  
  
  /* step 4 send reset to the device */
  mx1_IDE_reset();

  /* init ok */
  return 0;
}

static void mx1_IDE_info(void)
     /* permit to retrieve information in hard drive data (ident_command) */
{
  int i = 0;
  int j = 0;
 
  /* search specific parameters in hard drive info */
  mx1_ide_dev_id.ide_log_cyl       = mx1_ide_dev_id.ide_ident[1]; 
  mx1_ide_dev_id.ide_log_head      = mx1_ide_dev_id.ide_ident[3];
  mx1_ide_dev_id.ide_sec_trk       = mx1_ide_dev_id.ide_ident[6];
  mx1_ide_dev_id.ide_sec_count     = (unsigned int)mx1_ide_dev_id.ide_ident[58];
  mx1_ide_dev_id.ide_sec_count     = mx1_ide_dev_id.ide_sec_count << 16;
  mx1_ide_dev_id.ide_sec_count     = mx1_ide_dev_id.ide_sec_count + (unsigned int)mx1_ide_dev_id.ide_ident[57];
  mx1_ide_dev_id.ide_sec_lba_count = (unsigned int)mx1_ide_dev_id.ide_ident[61];
  mx1_ide_dev_id.ide_sec_lba_count = mx1_ide_dev_id.ide_sec_lba_count << 16;
  mx1_ide_dev_id.ide_sec_lba_count = mx1_ide_dev_id.ide_sec_lba_count + (unsigned int)mx1_ide_dev_id.ide_ident[60];
  for(i = 10; i<20; i++){
    mx1_ide_dev_id.ide_serial[j+1] = (char)mx1_ide_dev_id.ide_ident[i];
    mx1_ide_dev_id.ide_serial[j] = (char)(mx1_ide_dev_id.ide_ident[i] >> 8);
    j=j+2;
  }
  j = 0;
  for(i = 23; i<27; i++) {
    mx1_ide_dev_id.ide_firmware[j+1] = (char)mx1_ide_dev_id.ide_ident[i];
    mx1_ide_dev_id.ide_firmware[j] = (char)(mx1_ide_dev_id.ide_ident[i] >> 8);
    j=j+2;
  }
  j = 0;
  for(i = 27; i<47; i++) {
    mx1_ide_dev_id.ide_name[j+1] = (char)mx1_ide_dev_id.ide_ident[i];
    mx1_ide_dev_id.ide_name[j] = (char)(mx1_ide_dev_id.ide_ident[i] >> 8);
    j=j+2;
  }
  mx1_ide_dev_id.ide_name[40]= '\0';
  mx1_ide_dev_id.ide_serial[20] = '\0';
  mx1_ide_dev_id.ide_firmware[8] = '\0';
}

static void mx1_IDE_display_ident() 
     /* display hard drive informations */
{
  long capa;
  capa = (long)mx1_ide_dev_id.ide_sec_lba_count;
  capa = capa * 512 / 1000000000;
  printk(KERN_INFO "\n%s\n",mx1_ide_dev_id.ide_name);
  printk(KERN_INFO "serial :%s\n",mx1_ide_dev_id.ide_serial);
  printk(KERN_INFO "firmware :%s\n",mx1_ide_dev_id.ide_firmware);
  printk(KERN_INFO "%d logical cylinders (CHS)\n", mx1_ide_dev_id.ide_log_cyl);
  printk(KERN_INFO "%d logical heads (CHS)\n", mx1_ide_dev_id.ide_log_head);
  printk(KERN_INFO "%d sectors/track (CHS)\n", mx1_ide_dev_id.ide_sec_trk);
  printk(KERN_INFO "%i sectors, %ld GB (LBA)\n",mx1_ide_dev_id.ide_sec_lba_count, capa);
}

/******************************************************************************/
/*                                                                            */
/*                      IDE TASKLET                                           */
/*                                                                            */
/******************************************************************************/
static void mx1_IDE_errtask()
     /* error managment tasklet */
{
  /* TO DO, implement retry loop command */
  printk("ERROR TASKLET \n");
}

static void mx1_IDE_wtask()
     /* write operation tasklet */
{
  /* TO DO, verify if kernel presents 1 or 2 sector (and not more) */
  /* WRITE DMA transfers */
  
  unsigned long int_flags;
  save_flags(int_flags);
  cli();    

#ifdef DMA_MODE
	disable_dma(mx1ide_dma_chan_wr);		/* ensure DMA channel is disabled */
	set_dma_addr(mx1ide_dma_chan_wr, virt_to_phys((void *) mx1_ide_dev->cmd.buffer)); /* set DMA source address */
#endif

  *(volatile P_U32) (PTA_IMR_) |= 0x40000;  /* active interrupt for GPIO pin 18 */
  if(mx1_ide_dev->cmd.nsect == 1) { /* 1 writed sector 1*512 */ 
	    IDE_WRITE_CMD;
    	WAIT_BSY_DRQ;
#ifdef DMA_MODE
		set_dma_count(mx1ide_dma_chan_wr, MX1_IDE_HARDSECT_SIZE); /* set number of bytes to exchange */
#else
	    outsw_p(IDE_DATA_REG, mx1_ide_dev->cmd.buffer, 256);
#endif
  }
  if(mx1_ide_dev->cmd.nsect == 2) { /* 2 writed sector 2*512 */
    	IDE_WRITE_MULT_CMD;
    	WAIT_BSY_DRQ;
#ifdef DMA_MODE
		set_dma_count(mx1ide_dma_chan_wr, 2 * MX1_IDE_HARDSECT_SIZE); /* set number of bytes to exchange */
#else
    	outsw_p(IDE_DATA_REG, mx1_ide_dev->cmd.buffer, 512);
#endif
  }

#ifdef DMA_MODE
	enable_dma(mx1ide_dma_chan_wr);	/* start the DMA transfer */
#endif

	restore_flags(int_flags);
}

static void mx1_IDE_rtask()
     /* read operation tasklet */
{
  /* TO DO, verify if kernel presents 1 or 2 sector (and not more) */
  /* READ DMA transfers */
  unsigned long flags;
  unsigned long int_flags;

  if (mx1_ide_dev->cmd.command == IDE_READ)  {
    save_flags(int_flags);
    cli();
    if(mx1_ide_dev->cmd.nsect == 1)	{
#ifdef DMA_MODE
		disable_dma(mx1ide_dma_chan_rd);		/* ensure DMA channel is disabled */
		set_dma_addr(mx1ide_dma_chan_rd, virt_to_phys((void *) mx1_ide_dev->cmd.buffer)); /* set DMA source address */
		set_dma_count(mx1ide_dma_chan_rd, MX1_IDE_HARDSECT_SIZE); /* set number of bytes to exchange */
#else
		insw_p(IDE_DATA_REG,mx1_ide_dev->cmd.buffer,256); /* 1 readed sector 1*512 */
#endif
	}
    if(mx1_ide_dev->cmd.nsect == 2)	{
#ifdef DMA_MODE
		disable_dma(mx1ide_dma_chan_rd);		/* ensure DMA channel is disabled */
		set_dma_addr(mx1ide_dma_chan_rd, virt_to_phys((void *) mx1_ide_dev->cmd.buffer)); /* set DMA source address */
		set_dma_count(mx1ide_dma_chan_rd, 2 * MX1_IDE_HARDSECT_SIZE); /* set number of bytes to exchange */
#else
		insw_p(IDE_DATA_REG,mx1_ide_dev->cmd.buffer,512); /* 2 readed sector 2*512 */ 
#endif
	}

#ifdef DMA_MODE
	enable_dma(mx1ide_dma_chan_rd); /* start the DMA transfer */ 
#endif

	restore_flags(int_flags);
  } 
#ifdef DMA_MODE
  if (mx1_ide_dev->cmd.command == IDE_WRITE)  {
	disable_dma(mx1ide_dma_chan_wr);
	del_timer(&mx1_ide_dev->timer);
	spin_lock_irqsave(&io_request_lock, flags);
	end_request(1);
	CLEAR_DEV_BUSY;
	if(! QUEUE_EMPTY) mx1_IDE_request(NULL);  
	spin_unlock_irqrestore(&io_request_lock, flags);
	}
#else
	del_timer(&mx1_ide_dev->timer);
	spin_lock_irqsave(&io_request_lock, flags);
	end_request(1);
	CLEAR_DEV_BUSY;
	if(! QUEUE_EMPTY) mx1_IDE_request(NULL);  
	spin_unlock_irqrestore(&io_request_lock, flags);
#endif
}

/******************************************************************************/
/*                                                                            */
/*                      IDE IRQ HANDLER                                       */
/*                                                                            */
/******************************************************************************/
static void mx1_IDE_interrupt(int irq, void *dev, struct pt_regs *regs)
     /* acknowledge hard drive interruption */
{
  U8 alt_status = 0;
  U8 status = 0;

  if(((*(volatile P_U32)(PTA_ISR_))&0x40000) != 0x40000) { /* verify if it's GPIO 18 interrupt */
  return;};

  alt_status = inb_p((volatile P_U8) IDE_ALT_STAT_REG);
  if((alt_status&0x80) == 0x80) {
    *(volatile P_U32)(PTA_ISR_) = 0x40000; /* verify if it's a valid hard drive interrupt (ALT STATUS REGISTER) */
  return;};
 
  if(!WAIT_INT) {
    *(volatile P_U32)(PTA_ISR_) = 0x40000; /* verify if it's a waited interrupt */
    return;
  }

  if(DEV_ERROR) {
    *(volatile P_U32)(PTA_ISR_) = 0x40000; /* verify if there is no error (ERROR flag) */
    return;
  }
  
  *(volatile P_U32) (PTA_IMR_) &=  ~0x40000;  /* desactive interrupt for GPIO pin 18 */

  status = inb_p((volatile P_U8) IDE_STAT_REG);
  CLEAR_WAIT_INT;
  *(volatile P_U32) PTA_ISR_ = 0x40000;
 
  if((status&0x1) == 0x1) {
    SET_DEV_ERROR;
    return;
  }

  tasklet_schedule(&mx1_IDE_read_tasklet);
}

static int mx1_IDE_IRQ_request()
     /* request IRQ and GPIO 18 */
{
  int result = 0;
  
  printk(KERN_INFO "Request MX1/L interrupt channel 11... ");
 
  /* request interrupt channel 11 */
  result = request_irq(MX1_IDE_IRQ, &mx1_IDE_interrupt, SA_SHIRQ, "IDE DRIVE",  &mx1_IDE_interrupt);
  
  outl(0x40000, (volatile P_U32) PTA_ISR_);      /* clear interrupt */
  *(volatile P_U32) (PTA_IMR_)   |=  0x40000;  /* active interrupt for GPIO pin 18 */
 
  if (result == 0) printk(KERN_INFO "success\n\n");else {
    printk(KERN_INFO "fail\n\n");
    mx1_IDE_make_clean();
    return -EBUSY;  /* IRQ unavailable */ 
  }
  /* ok */
  return result;
}

/******************************************************************************/
/*                                                                            */
/*                      IDE DMA HANDLERS                                      */
/*                                                                            */
/******************************************************************************/
#ifdef DMA_MODE
static void
mx1ide_dma_intr_handler_rd(void)
{
	unsigned long flags;

	disable_dma(mx1ide_dma_chan_rd);

	del_timer(&mx1_ide_dev->timer);
	spin_lock_irqsave(&io_request_lock, flags);
	end_request(1);
	CLEAR_DEV_BUSY;
	if(! QUEUE_EMPTY) mx1_IDE_request(NULL);  
	spin_unlock_irqrestore(&io_request_lock, flags);
}

static void
mx1ide_dma_intr_handler_wr(void)
{
	/* Nothing to do until error recovery */
}

static void
mx1ide_dma_err_handler_rd(int error_type)
{
	if(error_type == MX1_DMA_ERR_TE)
		printk(KERN_ERR	"\nHDD IDE DMA RD: TRANSFER ERROR\n");
	if(error_type == MX1_DMA_ERR_BO)
		printk(KERN_ERR	"\nHDD IDE DMA RD: BUFFER OVERFLOW\n");
}

static void
mx1ide_dma_err_handler_wr(int error_type)
{
	if(error_type == MX1_DMA_ERR_TE)
		printk(KERN_ERR	"\nHDD IDE DMA WR: TRANSFER ERROR\n");
	if(error_type == MX1_DMA_ERR_BO)
		printk(KERN_ERR	"\nHDD IDE DMA WR: BUFFER OVERFLOW\n");
}
#endif
/******************************************************************************/
/*                                                                            */
/*                      IDE TIMER HANDLER                                     */
/*                                                                            */
/******************************************************************************/
static void mx1_IDE_expires()
     /* the command timout function */
{
  /* TO DO, implement a timeout subroutine to relauch last hard drive command */
  printk("\nTIME OUT Command\n");
  CLEAR_WAIT_INT;
  CLEAR_DEV_ERROR;
  inb_p((volatile P_U8) IDE_STAT_REG);
  *(volatile P_U32) PTA_ISR_ = 0x40000;
  mx1_IDE_transtask();
}

/******************************************************************************/
/*                                                                            */
/*                      IDE OPERATIONS SECTION                                */
/*                                                                            */
/******************************************************************************/
static void mx1_IDE_wait_signal(int signals, int wait_cond)
     /* active wait on STATUS REGISTER for several bits */
{
  /* TODO infinite while loop, must add a control to break and launch a error subroutine */
  U8 temp = 0; 
  temp = inb_p((volatile P_U8) IDE_STAT_REG);
  temp &= ~signals;
  
  while(temp != wait_cond) {
    temp = inb_p((volatile P_U8) IDE_STAT_REG);
    temp &= ~signals;
    if(DEV_ERROR) { printk("Exit ERROR\n"); return; } 
  }
}

static int mx1_IDE_set_mode()
     /* switch hard drive in specific transmission mode */
{
  U8 temp = 0; 
  mx1_ide_dev->cmd.command = IDE_SET_FEAT;
  WAIT_BSY_DRDY;
  outb_p(0x1, (volatile P_U8) IDE_SEC_COUNT_REG); /* no IORDY signal, PIO4 */
  outb_p(0x3, (volatile P_U8) IDE_FEAT_REG);      /* select mode */
  IDE_SET_FEAT_CMD;
  WAIT_BSY;
  temp = inb_p((volatile P_U8) IDE_STAT_REG);
  temp &= ~0xfe;
  if (temp == 0x1) {
    SET_DEV_ERROR; 
    return -1;
  }
  return 0;
}

static int mx1_IDE_set_multiple()
     /* permit to use multiple read and write commands, assume 1024B logical sector */
{
  U8 temp = 0; 
  mx1_ide_dev->cmd.command = IDE_SET_MULT;
   WAIT_BSY_DRDY;
  
  outb_p(0x2, (volatile P_U8) IDE_SEC_COUNT_REG); /* 2 sectors in burst (2*512) */
  IDE_SET_MULT_CMD;
  WAIT_BSY;
  temp = inb_p((volatile P_U8) IDE_STAT_REG);
  temp &= ~0xfe;
  if (temp == 0x1) {
    SET_DEV_ERROR; 
    return -1;
  }
  return 0;
}

static int mx1_IDE_ident_drive(void)
     /* identify device, retrieve information from the hard drive, work without interruption */
{
  U8 temp = 0;
  mx1_ide_dev->cmd.command = IDE_IDENT_DEV;
   WAIT_BSY_DRDY;
   IDE_IDENT_DEV_CMD;	
  WAIT_BSY_DRQ;
  /* look status register */
  temp = inb_p((volatile P_U8) IDE_STAT_REG);
  temp &= ~0xfe;
  if (temp == 0x1) {
    SET_DEV_ERROR;
    return -1;
  }
  /* reading device parameters */
  insw_p(IDE_DATA_REG, mx1_ide_dev_id.ide_ident, 256);
  return 0;
}

static int mx1_IDE_exec_diag()
     /* device self diagnostic command */
{
  U8 temp = 0;
  mx1_ide_dev->cmd.command = IDE_DEV_DIAG;
   WAIT_BSY_DRDY;
   IDE_DEV_DIAG_CMD;
  WAIT_BSY;
  /* look status register */
  temp = inb_p((volatile P_U8) IDE_STAT_REG);
  temp &= ~0xfe;
  if (temp == 0x1) {
    SET_DEV_ERROR;
    return -1;
  }
  /* look error register */
  temp = inb_p((volatile P_U8) IDE_ERR_REG);
  if(temp == 0x1) return 0;
  if(temp == 0x2) return -2;

  if(temp == 0x3) return -3;
  if(temp == 0x4) return -4;
  if(temp == 0x5) return -5;
  return -6;
}

static void mx1_IDE_reset()
     /* send reset */
{
  
  *(volatile P_U32) EIM_CS5L_ &= ~0x1;            /* CS5 in manual mode to activate daughter board buffers */
  *(volatile P_U32) EIM_CS5L_ |= 0x2;             /* assert CS5 */
  
  *(volatile P_U32) PTB_DR_ &= ~(0x1<<17);
  set_current_state(TASK_INTERRUPTIBLE);          /* allow IRQ */ 
  schedule_timeout(0.7 * HZ);                     /* hold reset signal for 70 ms */
  *(volatile P_U32) PTB_DR_ |= (0x1<<17);
  
  *(volatile P_U32) EIM_CS5L_ &= ~0x2;            /* deassert CS5 */
  *(volatile P_U32) EIM_CS5L_ |= 0x1;             /* CS5 in normal operation mode */

}

/******************************************************************************/
/*                                                                            */
/*                      BLOCK DEVICE OPERATIONS                               */
/*                                                                            */
/******************************************************************************/
static void  mx1_IDE_block_init()
     /* general block device initialization */
{  
  int i = 0;
  
  /* initialize block device parameters */
  read_ahead[ide_major] = ide_readahead;                /* set read ahead */
  for(i=0 ; i < MX1_IDE_NPART; i++)                     /* set hard sector size */
    ide_hardsect_size[i] = MX1_IDE_HARDSECT_SIZE;
  hardsect_size[ide_major] = ide_hardsect_size;
  for(i=0; i < MX1_IDE_NPART; i++)                      /* set max sectors per request */
    ide_maxs[i] = MX1_IDE_MAXS;
  max_sectors[ide_major] = ide_maxs;
  memset(ide_blk_size, 0, sizeof(int) * MX1_IDE_NPART); /* start with zero-sized partitions : we'll fix this later */
  blk_size[ide_major] = ide_blk_size;
  for(i=0; i < MX1_IDE_NPART; i++)                      /* set logical block size */
    ide_blksize_size[i] = MX1_IDE_BLKSIZE_SIZE;
  blksize_size[ide_major] = ide_blksize_size;
  
  /* fix up the gendisk structure */
  mx1_ide_gendisk.flags      = &mx1_ide_gendisk_flags;
  mx1_ide_gendisk.major      = ide_major;                                 /* major number */
  ide_blk_size[0]            = mx1_ide_dev_id.ide_sec_lba_count/1024*512; /* partitions sizes in kB */
  mx1_ide_gendisk.sizes      = ide_blk_size;
  memset(ide_partitions, 0, MX1_IDE_NPART * sizeof(struct hd_struct));
  ide_partitions[0].nr_sects = mx1_ide_dev_id.ide_sec_lba_count;          /* sector number */
  mx1_ide_gendisk.part       = ide_partitions;
  mx1_ide_gendisk.nr_real    = 1;
  
  /* add ourselves to the global list */
  add_gendisk(&mx1_ide_gendisk);

  /* initialize device request queue */
  blk_init_queue(BLK_DEFAULT_QUEUE(ide_major), DEVICE_REQUEST); 

  /* read partition table */
  printk(KERN_INFO "Hard drive ");
  register_disk(&mx1_ide_gendisk, MKDEV(ide_major,0), MX1_IDE_NPART, &mx1_IDE_bdops, mx1_ide_dev_id.ide_sec_lba_count);
}

static int mx1_IDE_open (struct inode *inode, struct file *filp)
     /* increment device use counter */
{ 
  if (DEVICE_NR(inode->i_rdev) >= 1)
    return -ENODEV;
  MOD_INC_USE_COUNT;
  return 0;
}

static int mx1_IDE_release (struct inode *inode, struct file *filp)
     /* decrement device use counter, flush buffer when null (umount operations) */ 
{ 
  MOD_DEC_USE_COUNT;
  return 0;
}

static int mx1_IDE_revalidate (kdev_t dev, int maxusage)
     /* reread partition table after modification (fdisk operation) */
{
  int target;
  struct gendisk * gdev;
  int max_p;
  int start;
  int i;

  target = DEVICE_NR(dev);
  gdev = &mx1_ide_gendisk;

  if (DEV_BUSY || mx1_ide_dev->usage > maxusage) {
      return -EBUSY;
  }
  max_p = gdev->max_p;
  start = target << gdev->minor_shift;
  
  for (i=max_p - 1; i >=0 ; i--) {
    int minor = start + i;
    kdev_t devi = MKDEV(MAJOR_NR, minor);
    struct super_block *sb = get_super(devi); 
    sync_dev(devi);
    if (sb)
      invalidate_inodes(sb);
    invalidate_buffers(devi);
    gdev->part[minor].start_sect = 0;
    gdev->part[minor].nr_sects = 0;
  }
  grok_partitions(gdev, target, 1<<6, 63*255*(mx1_ide_dev_id.ide_sec_lba_count / 255 / 63));
  return 0;
}

static int mx1_IDE_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
     /* low level primitive to access the device (fdisk operation) */
{ 
  int dev;
  if ((!inode) || !(inode->i_rdev))
    return -EINVAL;
  dev = DEVICE_NR(inode->i_rdev);
  switch (cmd) {
  case HDIO_GETGEO: /* return hard drive geometry */
    {
      struct hd_geometry g; 
      struct hd_geometry *loc = (struct hd_geometry *) arg;
      g.heads = mx1_ide_dev_id.ide_log_head;
      g.sectors = mx1_ide_dev_id.ide_sec_trk;
      g.cylinders = mx1_ide_dev_id.ide_sec_lba_count / g.heads / g.sectors;
      g.start = ide_partitions[MINOR(inode->i_rdev)].start_sect;
      return copy_to_user(loc, &g, sizeof g) ? -EFAULT : 0; 
    }
  case BLKGETSIZE:         /* return device size */
    if (!arg)  return -EINVAL;
    return put_user(ide_partitions[MINOR(inode->i_rdev)].nr_sects,(long *) arg);
  case BLKRRPART:          /* re-read partition tables */
    if (!capable(CAP_SYS_ADMIN)) return -EACCES;
    return mx1_IDE_revalidate(inode->i_rdev, 1);
  case BLKROSET:           /* for ioctl we don't understand, let the block layer handle them */
  case BLKROGET:
  case BLKRASET:
  case BLKRAGET:
  case BLKFLSBUF:
  case BLKPG:
    return blk_ioctl(inode->i_rdev, cmd, arg);
  default:
    return -EINVAL;
  }
}

static void mx1_IDE_transtask()
{
  unsigned char head, cyl_high, cyl_low, sec_nb;

  mx1_ide_dev->timer.expires = jiffies + 20 * HZ; /* 2s timer for timeout operation */
  add_timer(&mx1_ide_dev->timer);  
  sec_nb = (unsigned char)(mx1_ide_dev->cmd.block & 0xff);
  cyl_low = (unsigned char)((mx1_ide_dev->cmd.block & 0xff00)>>8);
  cyl_high = (unsigned char)((mx1_ide_dev->cmd.block & 0xff0000)>>16);
  head = (unsigned char)((mx1_ide_dev->cmd.block & 0xf000000)>>24);
  WAIT_BSY_DRDY;
  head &= ~0x10;
  head |= 0xe0;
  outb_p(head, (volatile P_U8) IDE_HEAD_REG);  
  outb_p(cyl_high, (volatile P_U8) IDE_CYL_HIGH_REG);
  outb_p(cyl_low, (volatile P_U8) IDE_CYL_LOW_REG);
  outb_p(sec_nb, (volatile P_U8) IDE_SEC_NB_REG);
  outb_p(mx1_ide_dev->cmd.nsect, (volatile P_U8) IDE_SEC_COUNT_REG);
  
  SET_WAIT_INT;
 
  if(mx1_ide_dev->cmd.command == IDE_READ) {
    *(volatile P_U32) (PTA_IMR_) |=  0x40000;  /* active interrupt for GPIO pin 18 */
    if(mx1_ide_dev->cmd.nsect == 1) IDE_READ_CMD;
    if(mx1_ide_dev->cmd.nsect == 2) IDE_READ_MULT_CMD;
  }
  else {
    tasklet_schedule(&mx1_IDE_write_tasklet);
  }
}

static void mx1_IDE_request(request_queue_t *q) 
{
  unsigned int block, nsect;
  long flags;
  unsigned long int_flags;

  if(DEV_BUSY) return;
  
  while(1) {
    CLEAR_DEV_ERROR;
    INIT_REQUEST;             /* returns when queue is empty */	
    
    if (MINOR(CURRENT->rq_dev) < MX1_IDE_NPART && CURRENT_DEV < 1 && CURRENT->sector + CURRENT->nr_sectors <= ide_partitions[MINOR(CURRENT->rq_dev)].nr_sects) {
      save_flags(int_flags);
      cli();    
      spin_lock_irqsave(&mx1_ide_dev->lock_dev, flags);
      
      block = CURRENT->sector + ide_partitions[MINOR(CURRENT->rq_dev)].start_sect;
      nsect = CURRENT->nr_sectors;
      switch (CURRENT->cmd) {
      case READ:
      	mx1_ide_dev->cmd.command = IDE_READ;
	break;
      case WRITE:
	mx1_ide_dev->cmd.command = IDE_WRITE;
	break;
      default:
	spin_unlock_irqrestore(&mx1_ide_dev->lock_dev, flags);
	restore_flags(int_flags);
	end_request(0);
	break;
      }
      
      mx1_ide_dev->cmd.buffer = CURRENT->buffer;
      mx1_ide_dev->cmd.block = block;
      mx1_ide_dev->cmd.nsect = nsect;
      mx1_IDE_transtask();
      SET_DEV_BUSY;
      spin_unlock_irqrestore(&mx1_ide_dev->lock_dev, flags);
      restore_flags(int_flags);
      return;
    }
  }
}

static struct block_device_operations mx1_IDE_bdops = {
  open:               mx1_IDE_open,
  release:            mx1_IDE_release,
  ioctl:              mx1_IDE_ioctl,
  owner:              THIS_MODULE,
};


/******************************************************************************/
/*                                                                            */
/*                      MODULE INITIALISATION FUNCTION                        */
/*                                                                            */
/******************************************************************************/
static int __init mx1_IDE_init(void)
{
  /* function result */
	int result = 0;

#ifdef DMA_MODE
	mx1ide_dma_chan_rd = -1;
	mx1ide_dma_chan_wr = -1;
#endif

  /* print some informations */
  printk(KERN_INFO "\niMX1/L IDE module ALPHA v 2.0\n");
  printk(KERN_INFO "2005 FREESCALE SEMICONDUCTOR\n\n");
  
  /* step 1 initialize internal IDE drive representation */
  mx1_ide_dev = kmalloc(sizeof(ide_dev), GFP_KERNEL);       /* allocate memory for ide structure */
  memset(mx1_ide_dev, 0, sizeof(ide_dev));                  /* 0 fill mx1_ide_dev structure */
  spin_lock_init(&mx1_ide_dev->lock_dev);                   /* initialize device spin lock */
  init_timer(&mx1_ide_dev->timer);                          /* initialize device timer */ 
  (mx1_ide_dev->timer).function = mx1_IDE_expires;          /* timer handler function */ 
  
  /* step 2 GPIO 17, EIM and memory initialization */
  result = mx1_IDE_hard_init();
  if (result != 0) return result;
  
  /* step 4 configure hard drive */
  result  = mx1_IDE_set_mode();                              /* disable IORDY signal */
  result += mx1_IDE_set_multiple();                          /* activate multiple data transfert */
  result += mx1_IDE_ident_drive();                           /* get hard drive info */
  mx1_IDE_info();
  if (result != 0) {
    mx1_IDE_make_clean();
    return result;
  }
  
  /* step 5 perform self-test and  print hard drive name, firmware, serial, capacity */
  mx1_IDE_display_ident();
  printk(KERN_INFO "Executing self-diagnostic test... ");
  result = mx1_IDE_exec_diag();
  if( result != 0 ) { 
    printk(KERN_INFO "fail !\n\n");
    mx1_IDE_make_clean();
    return result;
  }
  else printk(KERN_INFO "ok\n\n");
  
  /* step 3 request IRQ and GPIO 18 */
  result = mx1_IDE_IRQ_request();
  if (result != 0) return result;
  
  /* step 6 put major number for IDE drive and accept a dynamic number, no kernel devfs */
  result = devfs_register_blkdev(major, DEVICE_NAME, &mx1_IDE_bdops);
  if (result < 0) {
    printk(KERN_WARNING "IDE WARNING: unable to get major %d for IDE device !\n", major);
    free_irq(MX1_IDE_IRQ, &mx1_IDE_interrupt);
    mx1_IDE_make_clean();
    return result;
  }
  if ( major == 0 ) major = result;        /* dynamic */
  ide_major = major;

#ifdef DMA_MODE  
	/* request 2 DMA channels for RD and WR transfer */
	for (mx1ide_dma_chan_rd = 0; mx1ide_dma_chan_rd < MAX_DMA_CHANNELS;
	     mx1ide_dma_chan_rd++) {
		if (!request_dma(mx1ide_dma_chan_rd, "HDD IDE RD"))
			break;
	}

	for (mx1ide_dma_chan_wr = mx1ide_dma_chan_rd; mx1ide_dma_chan_wr < MAX_DMA_CHANNELS;
	     mx1ide_dma_chan_wr++) {
		if (!request_dma(mx1ide_dma_chan_wr, "HDD IDE WR"))
			break;
	}

	if (mx1ide_dma_chan_rd == MAX_DMA_CHANNELS) {
		printk(KERN_ERR "HDD IDE ERROR: could not allocate DMA for RD\n");
		mx1ide_dma_chan_rd = -1;
		mx1_IDE_make_clean();
		return -1;
	}

	if (mx1ide_dma_chan_wr == MAX_DMA_CHANNELS) {
		printk(KERN_ERR "HDD IDE ERROR: could not allocate DMA for WR\n");
		mx1ide_dma_chan_wr = -1;
		mx1_IDE_make_clean();
		return -1;
	}

		/* Read DMA channel settings */
	mx1_dma_chan_reg_out(DMA_SAR, mx1ide_dma_chan_rd, 0x16000010);	/* HDD Data Register */
	mx1_dma_chan_reg_out(DMA_BLR, mx1ide_dma_chan_rd, 16);	/* burst length : 16 bytes */
	mx1_dma_chan_reg_out(DMA_CCR, mx1ide_dma_chan_rd, 0x0820);	/*Lin-32bit Dest - FIFO-16bit Src */
	mx1_dma_chan_reg_out(DMA_RTOR, mx1ide_dma_chan_rd, 0);	/* burst timeout, not used */
	mx1_dma_chan_reg_out(DMA_BUCR, mx1ide_dma_chan_rd, 0);	/* bus utilization, not used */
	set_dma_mode(mx1ide_dma_chan_rd, DMA_MODE_READ);	/*this is for set_dma_addr() */
	mx1ads_request_dma_intr(mx1ide_dma_chan_rd,
				(callback_t) mx1ide_dma_intr_handler_rd,
				(err_callback_t) mx1ide_dma_err_handler_rd);

		/* Write DMA channel settings */
	mx1_dma_chan_reg_out(DMA_DAR, mx1ide_dma_chan_wr, 0x16000010);	/* HDD Data Register */
	mx1_dma_chan_reg_out(DMA_BLR, mx1ide_dma_chan_wr, 16);	/* burst length : 16 bytes */
	mx1_dma_chan_reg_out(DMA_CCR, mx1ide_dma_chan_wr, 0x2080);	/*FIFO-16bit Dest - Lin-32bit Src */
	mx1_dma_chan_reg_out(DMA_RTOR, mx1ide_dma_chan_wr, 0);	/* burst timeout, not used */
	mx1_dma_chan_reg_out(DMA_BUCR, mx1ide_dma_chan_wr, 0);	/* bus utilization, not used */
	set_dma_mode(mx1ide_dma_chan_wr, DMA_MODE_WRITE);	/*this is for set_dma_addr() */
	mx1ads_request_dma_intr(mx1ide_dma_chan_wr,
				(callback_t) NULL,//mx1ide_dma_intr_handler_wr,
				(err_callback_t) mx1ide_dma_err_handler_wr);
#endif

  /* step 7 block device initialization */
  mx1_IDE_block_init();

  /* LED2 turn on */
  LED2_ON;

  return 0; /* succeed */ 
}

/******************************************************************************/
/*                                                                            */
/*                      MODULE SHUTDOWN FUNCTION                              */
/*                                                                            */
/******************************************************************************/
static void __exit mx1_IDE_exit(void)
{
  int i = 0;

  /* flush all pending operations */
  flush_scheduled_tasks();

  /* clean major number device */
  devfs_unregister_blkdev(ide_major, DEVICE_NAME);

  /* flush device */
  for( i=0; i < MX1_IDE_NPART; i++ )
    fsync_dev(MKDEV(ide_major, i)); 

  /* clean I/O request queue */
  blk_cleanup_queue(BLK_DEFAULT_QUEUE(ide_major));

  /* stop internal ide structure */
  del_timer(&mx1_ide_dev->timer);
  spin_lock(&mx1_ide_dev->lock_dev);
  mx1_ide_dev->usage++;
  spin_unlock(&mx1_ide_dev->lock_dev);  
  /* clean internal ide structure */
  kfree(mx1_ide_dev);

  /* clean block device parameters */
  blk_size[ide_major] = NULL;
  blksize_size[ide_major] = NULL;
  hardsect_size[ide_major] = NULL;
  max_sectors[ide_major] = NULL;
  read_ahead[ide_major] = 0;
  
  /* clean IRQ 11 */
  free_irq(MX1_IDE_IRQ, &mx1_IDE_interrupt);

  /* clean gendisk structure */
  del_gendisk(&mx1_ide_gendisk);

#ifdef DMA_MODE
  /* free DMA channels */
	if (mx1ide_dma_chan_rd != -1) {
		mx1ads_free_dma_intr(mx1ide_dma_chan_rd);
		free_dma(mx1ide_dma_chan_rd);
	}
	if (mx1ide_dma_chan_wr != -1) {
		mx1ads_free_dma_intr(mx1ide_dma_chan_wr);
		free_dma(mx1ide_dma_chan_wr);
	}
#endif

  /* LED2 turns off */
  LED2_OFF;

  /* clean memory E/S */
  mx1_IDE_free();
}


/******************************************************************************/
/*                                                                            */
/*                      MODULE START/STOP FUNCTIONS                           */
/*                                                                            */
/******************************************************************************/
module_init(mx1_IDE_init);
module_exit(mx1_IDE_exit);



/* module information */
MODULE_AUTHOR ("Freescale Semiconductor");
MODULE_DESCRIPTION ("iMX1/L IDE module v 2.0");
MODULE_LICENSE("GPL");
